.shadow_widget {
  background-color: #5D7555;
}

.parent_widget {
  background-color: #1F4151;
}

.child_widget {
  background-color: #421F51;
}

.selected_widget {
  background-color: #83A67F;
}

.error_widget {
  background-color: #802020;
}

.dirty_widget {
  background-color: #4f3e22;
}

.alert_widget {
  background-color: #454555;
}

.centre_widget label {
  font-weight: bold;
  font-size: large;
}

.widget {
  background-color: #454545;
  color: white;
}

.error_top {
  font-size: large;
  font-weight: bold;
}

.error_sub {
  font-size: small;
}

.column_name {
  font-weight: bold;
}

.column_caption {
  font-size: small;
  font-style: italic;
}

.column_body {
  background-color: #353535;
}

.info {
  font-family: mono;
}

#rowview-frame {
  border-radius: 0px;
  border-bottom-width: 0px;
}

#rowview-frame.top {
  border-top-left-radius: 8px;
  border-top-right-radius: 8px;
}

#rowview-frame.bottom {
  border-bottom-left-radius: 8px;
  border-bottom-right-radius: 8px;
}

.left {
  border-top-right-radius: 0px;
  border-bottom-right-radius: 0px;
}

.right {
  border-top-left-radius: 0px;
  border-bottom-left-radius: 0px;
}

.tool-name {
  margin-top: 1px;
  margin-bottom: 1px;
  margin-left: 4px;
  margin-right: 4px;
}

.workspacedefs-text {
    background-color: #1d1d1d;
}

.toolkitgroupview {
    background-color: #1d1d1d;
}

.toolkitgroupview .view {
    background-color: #1d1d1d;
}

.toolkitgroupview-item {
  margin-top: -4px;
  margin-bottom: -4px;
}

.toolkitgroupview-sep {
  margin-top: -9px;
  margin-bottom: -10px;
}

imagedisplay {
  transition-property: outline-offset, outline-color;
  transition-duration: 300ms;
  animation-timing-function: ease-in-out;
  border-radius: 7px;
  outline-offset: 4px;
  outline: 2px solid transparent;
}

imagedisplay:focus:focus-visible {
  outline-offset: -3px;
  outline-color: @theme_selected_bg_color;
}
