// programming window

/*

	Copyright (C) 1991-2003 The National Gallery
	Copyright (C) 2004-2023 libvips.org

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

 */

#ifndef __PROGRAM_H
#define __PROGRAM_H

#define PROGRAM_TYPE (program_get_type())

#define PROGRAM NIP4_PROGRAM
#define IS_PROGRAM NIP4_IS_PROGRAM

G_DECLARE_FINAL_TYPE(Program, program, NIP4, PROGRAM, GtkApplicationWindow)

Program *program_new(App *app, Toolkitgroup *kitg);

#endif /* __PROGRAM_H */
