// shims for watch.h, until we get it in

#define NIP4_PATH "$HOME/.nip4-0.1.0"
#define AUTO_WS_SAVE (TRUE)
#define PRINT_CARTESIAN (FALSE)
#define LINELENGTH (80)
#define MAX_HEAPSIZE (10000000)
#define WORKSPACE_LPANE_OPEN (FALSE)
#define WORKSPACE_LPANE_POSITION (200)
#define WORKSPACE_RPANE_OPEN (FALSE)
#define WORKSPACE_RPANE_POSITION (400)
#define PAINTBOX_MAX_UNDO (10)
#define POPUP_NEW_ROWS (FALSE)
#define CALC_RECOMP_REGION (TRUE)
#define CALC_RELOAD (FALSE)
#define PATH_SEARCH path_search_default
#define PATH_START path_start_default
#define PATH_TMP path_tmp_default
#define DISPLAY_CONVERSION (TRUE)
